/*-
 * Copyright 2003-2005 Colin Percival
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted providing that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
// This file is a nearly copy of bsdiff.c from the
// bsdiff-4.3 distribution; the primary differences being how the
// input and output data are read, data post processing,
// search function modification and the error handling.

#define _CRT_SECURE_NO_WARNINGS
#include <sys/types.h>
#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include <Alloc.h>
#include <7zFile.h>
#include <7zVersion.h>
#include <LzmaDec.h>
#include <LzmaEnc.h>

#define SUFSORT_MOD					// Change suffix sorting algorithm from Qsufsort to Divsufsort
//#define ZLIB_MOD					// Change compression algorithm
#define SEARCH_RECURSIVE			// Change recursion of Search function to Iteration
//#define MAX_MATCH_SIZE			// define ( MAX_MATCH_SIZE or CONST_MEMORY_USAGE ) or ( none of them ); use max_match memory for old file at patch side
#define CONST_MEMORY_USAGE 16384		// patch in m+O(1); m=size of new file; use only const memory for old file at patch side;
#define PATCH_FILE_FORMAT_MOD       // no accumulation of diff and extra in db and eb; immediate write; also write all 3 parts of control stmt at same time
#define MULTI_THREADING 1            // only with #define CONST_MEMORY_USAGE or #define MAX_MATCH_SIZE
#define TIME_LIMIT_CHECK 300

/* Take care :
1) Use either (MAX_MATCH_SIZE or CONST_MEMORY_USAGE) or (none of both).
2.) PATCH_FILE_FORMAT_MOD may or may not be used, independently of everything else.
3.) MULTI_THREADING can be used only with (CONST_MEMORY_USAGE or MAX_MATCH_SIZE).
*/
#ifdef TIME_LIMIT_CHECK
long long outer_count = 0;
char ts1[256];
void get_time_stamp(void)
{
	struct timeval tv;
	int sec, msec;

	gettimeofday(&tv, NULL);
	sec = (int)tv.tv_sec;
	msec = (int)(tv.tv_usec / 1000);
	snprintf(ts1, 256, "%06d.%03d", sec % 100000, msec);
}
#endif
#ifdef SUFSORT_MOD
//supporting only 32 bit divsufsort for now.
#include <divsufsort.h>
#endif

#define MIN(x, y) (((x) < (y)) ? (x) : (y))

#ifdef MULTI_THREADING

struct data_thread {
	unsigned num_threads;
	off_t size_thread;

	int fd;
	int rc;
	u_char *old;
	u_char **new;
	off_t oldsize;
	off_t newsize;

	saidx_t *I;
	u_char buf[8];
	u_char header[16];
	u_char buf2[32];

	off_t lenn;
	off_t lenn2;

	FILE * pf;
	int bz2err;
	FILE * pfbz2;
};

struct data_thread data;

int Function(int);


#endif

static off_t matchlen(u_char *old, off_t oldsize, u_char *new, off_t newsize)
{
	off_t i;
	for (i = 0; (i < oldsize) && (i < newsize); i++)
		if (old[i] != new[i])
			break;
	return i;
}

static off_t search(saidx_t *I, u_char *old, off_t oldsize,
		u_char *new, off_t newsize, off_t st, off_t en, off_t *pos)
{
	off_t x, y;
	while (en - st >= 2) {
		x = st + (en - st) / 2;
		if (memcmp(old + I[x], new, MIN(oldsize - I[x], newsize)) < 0)
			st = x;
		else
			en = x;
	}

	x = matchlen(old + I[st], oldsize - I[st], new, newsize);
	y = matchlen(old + I[en], oldsize - I[en], new, newsize);

	if (x > y) {
		*pos = I[st];
		return x;
	} else {
		*pos = I[en];
		return y;
	}
}

static void offtout(off_t x, u_char *buf)
{
	off_t y;

	if (x < 0)
		y = -x;
	else
		y = x;

	buf[0] = y % 256;
	y -= buf[0];
	y = y / 256;
	buf[1] = y % 256;
	y -= buf[1];
	y = y / 256;
	buf[2] = y % 256;
	y -= buf[2];
	y = y / 256;
	buf[3] = y % 256;
	y -= buf[3];
	y = y / 256;
	buf[4] = y % 256;
	y -= buf[4];
	y = y / 256;
	buf[5] = y % 256;
	y -= buf[5];
	y = y / 256;
	buf[6] = y % 256;
	y -= buf[6];
	y = y / 256;
	buf[7] = y % 256;

	if (x < 0)
		buf[7] |= 0x80;
}

int create_patch(int argc, const char *argv[], int offset_oldscore)
{
	data.num_threads = MULTI_THREADING;
	data.new = (u_char **)malloc(sizeof(u_char *)*data.num_threads);
	if (argc != 4)
		errx(1, "usage: %s oldfile newfile patchfile\n", argv[0]);

	/* Allocate oldsize+1 bytes instead of oldsize bytes to ensure
		that we never try to malloc(0) and get a NULL pointer */
	if (((data.fd = open(argv[1], O_RDONLY, 0)) < 0) ||
			((data.oldsize = lseek(data.fd, 0, SEEK_END)) == -1) ||
			((data.old = malloc(data.oldsize + 1)) == NULL) ||
			(lseek(data.fd, 0, SEEK_SET) != 0) ||
			(read(data.fd, data.old, data.oldsize) != data.oldsize) ||
			(close(data.fd) == -1))
		err(1, "%s", argv[1]);

	data.I = malloc((data.oldsize + 1) * sizeof(saidx_t));
	divsufsort(data.old, data.I, data.oldsize);

	/* Allocate newsize+1 bytes instead of newsize bytes to ensure
		that we never try to malloc(0) and get a NULL pointer */
	if (((data.fd = open(argv[2], O_RDONLY, 0)) < 0) ||
			((data.newsize = lseek(data.fd, 0, SEEK_END)) == -1) ||
			(lseek(data.fd, 0, SEEK_SET) != 0))
		err(1, "%s", argv[2]);
	data.size_thread = (data.newsize / data.num_threads);

	unsigned int j;
	for (j = 0; j < data.num_threads; ++j) {
		if (j != data.num_threads - 1) {
			if (((data.new[j] = (u_char *)malloc(sizeof(u_char) * (data.size_thread + 1))) == NULL) ||
					(lseek(data.fd, 0, SEEK_CUR) != j * data.size_thread) ||
					(read(data.fd, data.new[j], data.size_thread) != data.size_thread))
				err(1, "%s", argv[2]);
		} else {
			if (((data.new[j] = (u_char *)malloc(sizeof(u_char) * (data.newsize - (j * data.size_thread) + 1))) == NULL) ||
					(lseek(data.fd, 0, SEEK_CUR) != j * data.size_thread) ||
					(read(data.fd, data.new[j], data.newsize - (j * data.size_thread)) != data.newsize - (j * data.size_thread)))
				err(1, "here %s", argv[2]);
		}
	}

	if ((close(data.fd) == -1))
		err(1, "%s", argv[2]);

	/* Create the patch file */
	if ((data.pf = fopen("temp_patch", "w")) == NULL)
		err(1, "%s", "temp_patch");

	/* Header is
		0	8	 "BSDIFF40"
		8	8	length of bzip2ed ctrl block
		16	8	length of bzip2ed diff block
		24	8	length of new file */
	/* File is
		0	32	Header
		32	??	Bzip2ed ctrl block
		??	??	Bzip2ed diff block
		??	??	Bzip2ed extra block */
	memcpy(data.header, "SSDIFF40", 8);

	offtout(data.newsize, data.header + 8);
	if (fwrite(data.header, 16, 1, data.pf) != 1)
		err(1, "fwrite(%s)", "temp_patch");

	/* Compute the differences, writing ctrl as we go */
	data.pfbz2 = data.pf;
	//if ((data.pfbz2 = BZ2_bzWriteOpen(&data.bz2err, data.pf, 9, 0, 0)) == NULL)
		//errx(1, "BZ2_bzWriteOpen, bz2err = %d", data.bz2err);


	//BZ2_bzWriteClose(&data.bz2err, data.pfbz2, 0, NULL, NULL);
	//if (data.bz2err != BZ_OK)
		//errx(1, "BZ2_bzWriteClose, bz2err = %d", data.bz2err);

	int ret = Function(offset_oldscore);
#ifdef TIME_LIMIT_CHECK
	if (ret != 0) {
		printf("bsdiff fails to create delta with offset score %d\n", offset_oldscore);
		printf("Old: [%s] -> New: [%s]\n", argv[1], argv[2]);
	}
#endif
	/* Seek to the beginning, write the header, and close the file */
	if (fseeko(data.pf, 0, SEEK_SET))
		err(1, "fseeko");

	if (fwrite(data.header, 16, 1, data.pf) != 1)
		err(1, "fwrite(%s)", "temp_patch");
	if (fclose(data.pf))
		err(1, "fclose");
	/* Free the memory we used */
	free(data.I);
	free(data.old);
	free(data.new);

	return ret;
}

int Function(int offset_oldscore)
{
	unsigned int thread_num = 0;
	off_t end;
	off_t scan = 0;
	end = data.newsize;
	int t1 = 0, t2 = 0;
	get_time_stamp();  //total time capturing
	t1 = atoi(ts1);

#ifdef PATCH_FILE_FORMAT_MOD
	u_char* db;
	u_char* eb;
	off_t dblen;
	off_t eblen;
#endif

	off_t pos;
	off_t len;
	off_t lastscan;
	off_t lastpos;
	off_t lastoffset;
	off_t oldscore;
	off_t scsc;
	off_t s;
	off_t Sf;
	off_t lenf;
	off_t Sb;
	off_t lenb;
	off_t overlap;
	off_t Ss;
	off_t lens;
	off_t i;

	pos = 0;
	len = 0;
	lastscan = 0;
	lastpos = 0;
	lastoffset = 0;
	while (scan < end) {
		oldscore = 0;
		for (scsc = scan += len; scan < end; scan++) {
			len = search(data.I, data.old, data.oldsize, data.new[thread_num] + scan, end - scan,
					len, data.oldsize, &pos); // Passing parameter as len instead of 0 for ramdisk.img etc taking long time

			for (; scsc < scan + len; scsc++)
				if ((scsc + lastoffset < data.oldsize) &&
						(data.old[scsc + lastoffset] == data.new[thread_num][scsc]))
					oldscore++;
#ifdef TIME_LIMIT_CHECK
			outer_count++;
			if (outer_count > 1000) {
				outer_count = 0;
				get_time_stamp();  //total time capturing
				t2 = atoi(ts1);
				//printf("\ntime diff = %d\n", (t2 - t1));
				if ((t2 - t1) > TIME_LIMIT_CHECK)
					return 1;
			}
#endif
			if (((len == oldscore) && (len != 0)) ||
					(len > oldscore + offset_oldscore))
				break;

			if ((scan + lastoffset < data.oldsize) &&
					(data.old[scan + lastoffset] == data.new[thread_num][scan]))
				oldscore--;
		};
		if ((len != oldscore) || (scan == end)) {
			s = 0;
			Sf = 0;
			lenf = 0;
			for (i = 0; (lastscan + i < scan) && (lastpos + i < data.oldsize); ) {
				if (data.old[lastpos + i] == data.new[thread_num][lastscan + i])
					s++;
				i++;
				if (s * 2 - i > Sf * 2 - lenf) {
					Sf = s;
					lenf = i;
				};
			};

			lenb = 0;
			if (scan < end) {
				s = 0;
				Sb = 0;
				for (i = 1; (scan >= lastscan + i) && (pos >= i); i++) {
					if (data.old[pos - i] == data.new[thread_num][scan - i])
						s++;
					if (s * 2 - i > Sb * 2 - lenb) {
						Sb = s;
						lenb = i;
					};
				};
			};

			if (lastscan + lenf > scan - lenb) {
				overlap = (lastscan + lenf) - (scan - lenb);
				s = 0;
				Ss = 0;
				lens = 0;
				for (i = 0; i < overlap; i++) {
					if (data.new[thread_num][lastscan + lenf - overlap + i] ==
							data.old[lastpos + lenf - overlap + i])
						s++;
					if (data.new[thread_num][scan - lenb + i] ==
							data.old[pos - lenb + i])
						s--;
					if (s > Ss) {
						Ss = s;
						lens = i + 1;
					};
				};

				lenf += lens - overlap;
				lenb -= lens;
			};

			if (((db = malloc(lenf + 1)) == NULL) ||
					((eb = malloc((scan - lenb) - (lastscan + lenf) + 1)) == NULL))
				err(1, NULL);

			for (i = 0; i < lenf; i++)
				db[i] = data.new[thread_num][lastscan + i] - data.old[lastpos + i];
			for (i = 0; i < (scan - lenb) - (lastscan + lenf); i++)
				eb[i] = data.new[thread_num][lastscan + lenf + i];
			dblen = lenf;
			eblen = (scan - lenb) - (lastscan + lenf);
			offtout(lenf, data.buf2);
			offtout((scan - lenb) - (lastscan + lenf), data.buf2 + 8);
			offtout(lastpos, data.buf2 + 16);
			offtout((data.size_thread * thread_num) + lastscan, data.buf2 + 24);
			fwrite(data.buf2, 1, 32, data.pf);
			//if (data.bz2err != BZ_OK)
				//errx(1, "fwrite, bz2err = %d", data.bz2err);

			fwrite(db, 1, dblen, data.pf);
			//if (data.bz2err != BZ_OK)
				//errx(1, "fwrite, bz2err = %d", data.bz2err);

			fwrite(eb, 1, eblen, data.pf);
			//if (data.bz2err != BZ_OK)
				//errx(1, "fwrite, bz2err = %d", data.bz2err);

			free(db);
			free(eb);

			lastscan = scan - lenb;
			lastpos = pos - lenb;
			lastoffset = pos - scan;
		};
	};
	return 0;
}

const char *kCantReadMessage = "Can not read input file";
const char *kCantWriteMessage = "Can not write output file";
const char *kCantAllocateMessage = "Can not allocate memory";
const char *kDataErrorMessage = "Data error";

static void *SzAlloc(void *p, size_t size)
{
	p = p;
	return MyAlloc(size);
}
static void SzFree(void *p, void *address)
{
	p = p;
	MyFree(address);
}
static ISzAlloc g_Alloc = { SzAlloc, SzFree };

int PrintError(char *buffer, const char *message, int buf_size)
{
	snprintf(buffer + strlen(buffer), buf_size - strlen(buffer),
			"\nError: %s\n", message);
	return 1;
}

int PrintErrorNumber(char *buffer, SRes val, int buf_size)
{
	snprintf(buffer + strlen(buffer), buf_size - strlen(buffer),
			"\nError code: %x\n", (unsigned)val);
	return 1;
}

int PrintUserError(char *buffer, int buf_size)
{
	return PrintError(buffer, "Incorrect command", buf_size);
}

#define IN_BUF_SIZE (1 << 16)
#define OUT_BUF_SIZE (1 << 16)


static SRes Encode(ISeqOutStream *outStream, ISeqInStream *inStream, UInt64 fileSize, char *rs)
{
	CLzmaEncHandle enc;
	SRes res;
	CLzmaEncProps props;

	rs = rs;

	enc = LzmaEnc_Create(&g_Alloc);
	if (enc == 0)
		return SZ_ERROR_MEM;

	LzmaEncProps_Init(&props);
	res = LzmaEnc_SetProps(enc, &props);

	if (res == SZ_OK) {
		Byte header[LZMA_PROPS_SIZE + 8];
		size_t headerSize = LZMA_PROPS_SIZE;
		int i;

		res = LzmaEnc_WriteProperties(enc, header, &headerSize);
		for (i = 0; i < 8; i++)
			header[headerSize++] = (Byte)(fileSize >> (8 * i));
		if (outStream->Write(outStream, header, headerSize) != headerSize)
			res = SZ_ERROR_WRITE;
		else {
			if (res == SZ_OK)
				res = LzmaEnc_Encode(enc, outStream, inStream, NULL, &g_Alloc, &g_Alloc);
		}
	}
	LzmaEnc_Destroy(enc, &g_Alloc, &g_Alloc);
	return res;
}

int main2(int numArgs, const char *args[], char *rs, int rs_size)
{
	CFileSeqInStream inStream;
	CFileOutStream outStream;
	int res;

	FileSeqInStream_CreateVTable(&inStream);
	File_Construct(&inStream.file);

	FileOutStream_CreateVTable(&outStream);
	File_Construct(&outStream.file);

	size_t t4 = sizeof(UInt32);
	size_t t8 = sizeof(UInt64);
	if (t4 != 4 || t8 != 8)
		return PrintError(rs, "Incorrect UInt32 or UInt64", rs_size);

	if (InFile_Open(&inStream.file, "temp_patch") != 0)
		return PrintError(rs, "Can not open input file", rs_size);


	if (OutFile_Open(&outStream.file, args[3]) != 0)
		return PrintError(rs, "Can not open output file", rs_size);


	UInt64 fileSize;
	File_GetLength(&inStream.file, &fileSize);
	res = Encode(&outStream.s, &inStream.s, fileSize, rs);

	File_Close(&outStream.file);
	File_Close(&inStream.file);

	if (res != SZ_OK) {
		if (res == SZ_ERROR_MEM)
			return PrintError(rs, kCantAllocateMessage, rs_size);
		else if (res == SZ_ERROR_DATA)
			return PrintError(rs, kDataErrorMessage, rs_size);
		else if (res == SZ_ERROR_WRITE)
			return PrintError(rs, kCantWriteMessage, rs_size);
		else if (res == SZ_ERROR_READ)
			return PrintError(rs, kCantReadMessage, rs_size);
		return PrintErrorNumber(rs, res, rs_size);
	}
	return 0;
}

int MY_CDECL main(int numArgs, const char *args[])
{
	char rs[800] = { 0 };
	if (numArgs != 4)
		errx(1, "ss_bsdiff Version 5.0\nUsage: ss_bsdiff oldfile newfile patchfile\n");

	int ret = create_patch(numArgs, args, 8);
#ifdef TIME_LIMIT_CHECK
	if (ret != 0) {
		printf("Trying with offset score 2\n");
		ret = create_patch(numArgs, args, 2);
	}
	if (ret != 0) {
		printf("Trying with offset score 0\n");
		ret = create_patch(numArgs, args, 0);
	}
	if (ret != 0)
		err(1, "bsdiff fails to create delta within timelimit");
#endif
	int res = main2(numArgs, args, rs, sizeof(rs));
	if (remove("temp_patch") < 0)
		printf("Failed to remove temp_patch\n");
	fputs(rs, stdout);
	return res;
}
