/*-
 * Copyright 2003-2005 Colin Percival
 * Copyright 2012 Matthew Endsley
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted providing that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Modifications are made in reimplementing suffix sort array generation
 * and how the data is read and written to.Iterative part replaced the
 * recursive implementation to avoid buffer overflow problems
 */
#define _CRT_SECURE_NO_WARNINGS
#include <err.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Alloc.h>
#include <7zFile.h>
#include <7zVersion.h>
#include <LzmaDec.h>
#include <LzmaEnc.h>

#include "ss_bspatch_common.h"

const char *kCantReadMessage = "Can not read input file";
const char *kCantWriteMessage = "Can not write output file";
const char *kCantAllocateMessage = "Can not allocate memory";
const char *kDataErrorMessage = "Data error";

int PrintError(char *buffer, const char *message, int buf_size)
{
	snprintf(buffer + strlen(buffer), buf_size - strlen(buffer),
			"\nError: %s\n", message);
	return 1;
}

int PrintErrorNumber(char *buffer, SRes val, int buf_size)
{
	snprintf(buffer + strlen(buffer), buf_size - strlen(buffer),
			"\nError code: %x\n", (unsigned)val);
	return 1;
}

int PrintUserError(char *buffer, int buf_size)
{
	return PrintError(buffer, "Incorrect command", buf_size);
}

int main2(int numArgs, const char *args[], char *rs, int rs_size)
{
	CFileSeqInStream inStream;
	CFileOutStream outStream;
	int res, fd;
	int encodeMode;
	unsigned char *buf_res = NULL;
	unsigned char *new_data;
	ssize_t new_size;

	FileSeqInStream_CreateVTable(&inStream);
	File_Construct(&inStream.file);

	FileOutStream_CreateVTable(&outStream);
	//File_Construct(&outStream.file);

	encodeMode = 0;

	size_t t4 = sizeof(UInt32);
	size_t t8 = sizeof(UInt64);
	if (t4 != 4 || t8 != 8)
		return PrintError(rs, "Incorrect UInt32 or UInt64", rs_size);

	if (InFile_Open(&inStream.file, args[3]) != 0)
		return PrintError(rs, "Can not open input file", rs_size);

	if (encodeMode) {
		UInt64 fileSize;
		File_GetLength(&inStream.file, &fileSize);
		//res = Encode(&outStream.s, &inStream.s, fileSize, rs);
	} else {
		UInt64 unpackSize, i;
		CLzmaDec state;
		unsigned char header[LZMA_PROPS_SIZE + 8];
		RINOK(SeqInStream_Read(&inStream.s, header, sizeof(header)));
		unpackSize = 0;
		for (i = 0; i < 8; i++)
			unpackSize += (UInt64)header[LZMA_PROPS_SIZE + i] << (i * 8);
		buf_res = (unsigned char *)malloc(unpackSize);
		if (!buf_res)
			return PrintError(rs, "Failed to allocate memory", rs_size);
		memset(buf_res, 0x0, unpackSize);
		LzmaDec_Construct(&state);
		RINOK(LzmaDec_Allocate(&state, header, LZMA_PROPS_SIZE, &g_Alloc));
		res = Decode2(&state, &outStream.s, &inStream.s, &unpackSize, buf_res);
		LzmaDec_Free(&state, &g_Alloc);
		File_Close(&inStream.file);
		if (apply_patch(args[1], buf_res, &new_data, &new_size) != 0) {
			if (new_data)
				free(new_data);
			return 1;
		}
		if (((fd = open(args[2], O_CREAT | O_TRUNC | O_WRONLY, 0666)) < 0) ||
				(write(fd, new_data, new_size) != new_size) || (close(fd) == -1))
			err(1, "%s", args[2]);
		if (res != SZ_OK) {
			free(new_data);
			free(buf_res);
			if (res == SZ_ERROR_MEM)
				return PrintError(rs, kCantAllocateMessage, rs_size);
			else if (res == SZ_ERROR_DATA)
				return PrintError(rs, kDataErrorMessage, rs_size);
			else if (res == SZ_ERROR_WRITE)
				return PrintError(rs, kCantWriteMessage, rs_size);
			else if (res == SZ_ERROR_READ)
				return PrintError(rs, kCantReadMessage, rs_size);
			return PrintErrorNumber(rs, res, rs_size);
		}
		free(new_data);
		free(buf_res);
		return 0;
	}
	return 1;
}

int main(int numArgs, const char *args[])
{
	char rs[800] = { 0 };
	if (numArgs != 4)
		errx(1, "ss_bspatch Version 1.0\nUsage: ss_bspatch oldfile newfile patchfile\n");
	int res = main2(numArgs, args, rs, sizeof(rs));
	fputs(rs, stdout);
	return res;
}
