/*-
 * Copyright 2003-2005 Colin Percival
 * Copyright 2012 Matthew Endsley
 * All rights reserved
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted providing that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *	notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *	notice, this list of conditions and the following disclaimer in the
 *	documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * Modifications are made in reimplementing suffix sort array generation
 * and how the data is read and written to.Iterative part replaced the
 * recursive implementation to avoid buffer overflow problems
 */
#include <string.h>
#include <err.h>
#include "SS_PatchDelta.h"
#include "fota_common.h"
#include "sha1.h"
#include "SS_Engine_Errors.h"
#include "SS_FSUpdate.h"

#include "ss_bspatch_common.h"
#include <Alloc.h>
#include <7zFile.h>
#include <7zVersion.h>
#include <LzmaDec.h>
#include <LzmaEnc.h>


int SS_ApplyBsdiff(char *oldfile, char *newfile, char *patch, SinkFn sink, void *token, sha1_ctx_t * ctx1)
{
	UInt64 unpackSize = 0;
	CFileSeqInStream inStream;
	ISeqOutStream outStream;
	unsigned char *buf_res = NULL;
	unsigned char *new_data = NULL;
	char buf[256];
	ssize_t new_size = 0;
	int result = E_SS_FAILURE;

	FileSeqInStream_CreateVTable(&inStream);
	File_Construct(&inStream.file);
	FileOutStream_CreateVTable((CFileOutStream *) & outStream);

	if (InFile_Open(&inStream.file, patch) != 0)
		return E_SS_FAILURE;

	UInt64 i;
	CLzmaDec state;
	unsigned char header[LZMA_PROPS_SIZE + 8];

	RINOK(SeqInStream_Read(&inStream.s, header, sizeof(header)));

	unpackSize = 0;
	for (i = 0; i < 8; i++)
		unpackSize += (UInt64) header[LZMA_PROPS_SIZE + i] << (i * 8);

	LzmaDec_Construct(&state);
	RINOK(LzmaDec_Allocate(&state, header, LZMA_PROPS_SIZE, &g_Alloc));

	//decompress the patch file into buf_res
	buf_res = (unsigned char *)SS_Malloc(unpackSize);
	if (!buf_res) {
		LOGE("Bad memory allocation\n");
		goto Cleanup;
	}
	result = Decode2(&state, &outStream, &inStream.s, &unpackSize, buf_res);

	LzmaDec_Free(&state, &g_Alloc);
	File_Close(&inStream.file);

	if (result != S_SS_SUCCESS) {
		LOGE("Error decompression failed with code : [%d]\n", result);
		goto Cleanup;
	}
	//apply patch using buffer decoded by Decode2
	result = apply_patch(oldfile, buf_res, &new_data, &new_size);
	if (result != S_SS_SUCCESS)
		goto Cleanup;

	result = (sink(new_data, new_size, token) < new_size) ? E_SS_FAILURE : S_SS_SUCCESS;
	if (result != S_SS_SUCCESS) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("short write of output: %d (%s)\n", errno, buf);
		goto Cleanup;
	}

	if (ctx1)
		sha1_update(ctx1, new_data, new_size);
Cleanup:
	if (new_data)
		SS_Free(new_data);
	if (buf_res)
		SS_Free(buf_res);
	File_Close(&inStream.file);//wgid: 27007
	return result;

}
