/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 *******************************************************************************
 * \file	SS_ImageUpdate.h
 *
 * \brief	UPI Image Update API
 *******************************************************************************
 */
#ifndef	__SS_IMAGEUPDATE_H__
#define	__SS_IMAGEUPDATE_H__

#include "SS_Engine_Update.h"
#include "SS_Engine_Errors.h"

#ifdef __cplusplus
extern "C" {
#endif                          /* __cplusplus */
/**
 *******************************************************************************
 * Get data from temporary storage.
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	pbBuffer		(out) Buffer to store the data
 * \param	dwBlockAddress	Location of data
 * \param	dwSize			Size of data
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_ReadBackupBlock(void *pbUserData, unsigned char *pbBuffer, SS_UINT32 dwBlockAddress, SS_UINT32 dwSize);

/**
 *******************************************************************************
 * Erase sector from temporary storage.<p>
 *
 * A default implementation is included. You must use the default implementation
 * if both of the following are true:
 * \li	<b>InstantFailSafe</b> is set to false
 * \li	Only Image updates are used OR the backup is stored in flash and not on
 *		the file system
 * <p>The address must be aligned to the sector size.
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	dwStartAddress	Sector address to erase
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_EraseBackupBlock(void *pbUserData, SS_UINT32 dwStartAddress);

/**
 *******************************************************************************
 * Write complete sector to temporary storage.<p>
 *
 * Must overwrite entire sector. To write part of a sector, see
 * \ref SS_WriteBackupPartOfBlock.
 *
 * \param	pbUserData			Optional opaque data-structure to pass to IPL
 *								functions
 * \param	dwBlockStartAddress	Address to which to write
 * \param	pbBuffer			Data to write
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_WriteBackupBlock(void *pbUserData, SS_UINT32 dwBlockStartAddress, unsigned char *pbBuffer);

/**
 *******************************************************************************
 * Write data to temporary storage.<p>
 *
 * Used to write part of a sector that has already been written to by
 * \ref SS_WriteBackupBlock. Must NOT overwrite entire sector.<p>
 *
 * A default implementation is included. You must use the default implementation
 * if both of the following are true:
 * \li	<b>InstantFailSafe</b> is set to false
 * \li	Only Image updates are used OR the backup is stored in flash and not on
 *		the file system
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	dwStartAddress	Address to which to write
 * \param	dwSize			Size of \a pbBuffer
 * \param	pbBuffer		Data to write
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_WriteBackupPartOfBlock(void *pbUserData,
			SS_UINT32 dwStartAddress, SS_UINT32 dwSize, unsigned char *pbBuffer);

/**
 *******************************************************************************
 * Write sector to flash.<p>
 *
 * Must overwrite a complete sector.
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	dwBlockAddress	Address to which to write
 * \param	pbBuffer		Data to write
 *
 * \return	S_SS_SUCCESS on success, E_SS_WRITE_ERROR if there is a write error,
 *			or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_WriteBlock(void *pbUserData, SS_UINT32 dwBlockAddress, unsigned char *pbBuffer);

/**
 *******************************************************************************
 * Read data from flash.<p>
 *
 * See \ref SS_ReadImageNewKey.
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	pbBuffer		(out) Buffer to store data
 * \param	dwStartAddress	Address from which to read
 * \param	dwSize			Size of data to read (must be less than a sector); 0
 *							must immediately return success
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_ReadImage(void *pbUserData, unsigned char *pbBuffer,        /* pointer to user buffer */
			SS_UINT32 dwStartAddress, /* memory address to read from */
			SS_UINT32 dwSize);        /* number of bytes to copy */

/**
 *******************************************************************************
 * Read data from flash.<p>
 *
 * As opposed to \ref SS_ReadImage, this function is intended to read already
 * updated data and is used for encrypted images. If no part of the image is
 * encrypted, simply call \ref SS_ReadImage.
 *
 * Used by the UPI when resuming an interrupted update.
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 * \param	pbBuffer		(out) Buffer to store data
 * \param	dwStartAddress	Address from which to read
 * \param	dwSize			Size of data to read (must be less than a sector); 0
 *							must immediately return success
 *
 * \return	S_SS_SUCCESS on success or an \ref SS_vRM_Errors.h error code
 *******************************************************************************
 */
	long SS_ReadImageNewKey(void *pbUserData, unsigned char *pbBuffer, SS_UINT32 dwStartAddress, SS_UINT32 dwSize);

/**
 *******************************************************************************
 * Get sector size.<p>
 *
 * \param	pbUserData		Optional opaque data-structure to pass to IPL
 *							functions
 *
 * \return	Sector size, in bytes
 *******************************************************************************
 */
	long SS_GetBlockSize(void *pbUserData);

#ifdef __cplusplus
}
#endif                          /* __cplusplus */
#endif                          /*      __SS_UPDATE_IMAGE_H__   */
