/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include "ua_types.h"
#include "sha1.h"
#include "SS_PatchDelta.h"
#include "fota_common.h"
#include "SS_Engine_Errors.h"

extern void *SS_Malloc(unsigned int size);

typedef struct {
	unsigned char *buffer;
	ssize_t size;
	ssize_t pos;
} MemorySinkInfo;

ssize_t ss_memorySink(unsigned char *data, ssize_t len, void *token)
{
	MemorySinkInfo *msi = (MemorySinkInfo *) token;
	if (msi->size - msi->pos < len)
		return -1;
	memcpy(msi->buffer + msi->pos, data, len);
	msi->pos += len;
	return len;
}

ssize_t ss_fileSink(unsigned char *data, ssize_t len, void *token)
{
	int ss_fd = *(int *)token;
	char buf[256];
	ssize_t done = 0;
	ssize_t wrote;
	while (done < (ssize_t) len) {
		wrote = write(ss_fd, data + done, len - done);
		if (wrote <= 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;	   // try again
		strerror_r(errno, buf, sizeof(buf));
			LOGE("error writing %d bytes: %s\n", (int)(len - done), buf);
			return done;
		}
		done += wrote;
	}
	return done;
}

// Take a string 'str' of 40 hex digits and parse it into the 20
// byte array 'digest'.  'str' may contain only the digest or be of
// the form "<digest>:<anything>".  Return 0 on success, -1 on any
// error.
int ParseSha1(const char *str, uint8_t * digest)
{
	int i;
	const char *ps = str;
	uint8_t *pd = digest;
	for (i = 0; i < SHA_DIGEST_SIZE * 2; ++i, ++ps) {
		int digit;
		if (*ps >= '0' && *ps <= '9')
			digit = *ps - '0';
		else if (*ps >= 'a' && *ps <= 'f')
			digit = *ps - 'a' + 10;
		else if (*ps >= 'A' && *ps <= 'F')
			digit = *ps - 'A' + 10;
		else
			return -1;
		if (i % 2 == 0) {
			*pd = digit << 4;
		} else {
			*pd |= digit;
			++pd;
		}
	}
	if (*ps != '\0')
		return -1;
	return 0;
}

//Function to find the start of gzipped part in compressed kernel
//This function will be removed during further optimization of kernel delta
int getOffset(char *zimage_path)
{
	char gzip_header[] = { 31, -117, 8 };	   //header value for gzip which needs to be checked
	char buf[4] = { 0, };
	int offset = -1;//wgid:14074

	FILE *f = fopen(zimage_path, "r");
	if (!f) {
		LOGE("Fopen failed for path %s\n", zimage_path);
		SS_SetUpgradeState(E_SS_OPENFILE_ONLYR);
		return -1;
	}
	fseek(f, 0, SEEK_SET);
	while (fread(buf, 1, 3, f) > 0) {
		if (gzip_header[0] == buf[0] && gzip_header[1] == buf[1] && gzip_header[2] == buf[2]) {
			LOGL(LOG_SSENGINE, "match for %d %d %d found at %d\n", buf[0], buf[1], buf[2], ftell(f) - 3);
			offset = ftell(f) - 3;//145599
			break;
		} else {
			if (fseek(f, -2, SEEK_CUR) < 0)
				LOGE("Failed to fseek\n");
		}
	}
	fclose(f);
	return offset;
}

int SS_LoadPartition(const char *filename, FileInfo * file)
{
	size_t read = 0;
	FILE *dev = NULL;
	int i;
	char buf[256];

	dev = fopen(filename, "rb");
	if (dev == NULL) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to open partition \"%s\": %s\n", filename, buf);
		return -1;
	}

	sha1_ctx_t sha_ctx;
	sha1_init(&sha_ctx);

	file->data = SS_Malloc(file->size);
	if (file->data) {
		read = fread(file->data, 1, file->size, dev);
		LOGL(LOG_SSENGINE, "Partition size read %d\n", read);
		sha1_update(&sha_ctx, file->data, read);
		file->size = read;
	}

	const uint8_t sha_final[SHA_DIGEST_SIZE] = { 0, };
	sha1_final(&sha_ctx, (uint32_t *) & sha_final);
	for (i = 0; i < SHA_DIGEST_SIZE; ++i)
		file->sha1[i] = sha_final[i];
	//LOGL(LOG_SSENGINE, "Final SHA of Source (%s)\n", sha_final);

	file->st.st_mode = 0644;
	file->st.st_uid = 0;
	file->st.st_gid = 0;
	fclose(dev);
	return 0;
}

//extern int write_data_to_blkdev(char* dev_name, int blk_start, int blk_cnt, char* data);

int SS_LoadFile(const char *filename, FileInfo * file)
{
	char buf[256];

	file->data = NULL;
	//LOGL(LOG_SSENGINE,"SS_LoadFile --- [File name %s]\n",filename);

	if (stat(filename, &file->st) != 0) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to stat \"%s\": %s\n", filename, buf);
		return -1;
	}

	file->size = file->st.st_size;
	file->data = SS_Malloc(file->size);
	if (!file->data) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to allocate memory for \"%s\": %s\n", filename, buf);
		return -1;
	}

	FILE *f = fopen(filename, "rb");
	if (f == NULL) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("failed to open \"%s\": %s\n", filename, buf);
		SS_Free(file->data);
		file->data = NULL;
		return -1;
	}

	ssize_t bytes_read = fread(file->data, 1, file->size, f);
	if (bytes_read != file->size) {
		LOGE("short read of \"%s\" (%ld bytes of %ld)\n", filename, (long)bytes_read, (long)file->size);
		SS_Free(file->data);
		file->data = NULL;
		fclose(f);
		return -1;
	}
	fclose(f);
	//LOGL(LOG_SSENGINE,"SS_LoadFile --- [bytes_read %d]\n",bytes_read);
	sha1(file->data, file->size, (uint32_t *) file->sha1);
	return 0;
}

extern int gvalid_session;
extern void create_dir(char *pathname, int mode);

/*!
 *********************************************************************************
 *					 SS_UpdateDeltaFS
 *********************************************************************************
 *
 * @brief
 *	This is used to apply patch for a file during delta FS upgrade
 *
 *
 *	@param
 *
 *	@return				0 - in case of success
 *						1 - in case of error during patch application
 *
 *********************************************************************************
 */

int SS_UpdateDeltaFS(const char *source_filename, const char *target_filename,
					 const char *source_sha1_str, const char *target_sha1_str, int patch_data_size)
{
	uint8_t target_sha1[SHA_DIGEST_SIZE] = { 0, };
	uint8_t source_sha1[SHA_DIGEST_SIZE] = { 0, };
	sha1_ctx_t ctx1;
	int output;
	int retry = 1;
	int use_backup = 0;
	char *outname = NULL;
	int backupsrc = -1;
	int result = 0;
	char buf[256];
	FileInfo source_file;
	FileInfo target_file;

	if (ParseSha1(target_sha1_str, target_sha1) != 0) {
		LOGE("failed to parse tgt-sha1 \"%s\"\n", target_sha1_str);
		return E_SS_FAILURE;
	}

	/*
	   if battery removed in between update gvalid_session becomes 0
	   need to check file integrity in that case
	 */
	if (0 == gvalid_session) {
		if (ParseSha1(source_sha1_str, source_sha1) != 0) {
			LOGE("failed to parse tgt-sha1 \"%s\"\n", source_sha1_str);
			return E_SS_FAILURE;
		}
		if (SS_LoadFile(source_filename, &source_file) == 0) {
			SS_Free(source_file.data);

			if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "Patch Can be applied\n");
			} else if (memcmp(source_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "Patch Already applied\n");
				return S_SS_SUCCESS;
			} else {
				//Check for backup file SHA
				LOGL(LOG_SSENGINE, "Source was currupted, Try loading from backup source\n");

				if (SS_LoadFile(target_filename, &target_file) == 0) {
					SS_Free(target_file.data);

					if (memcmp(target_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
						LOGL(LOG_SSENGINE, "Patch Already applied\n");
						return S_SS_SUCCESS;
					}
				}

				if (SS_LoadFile(SS_BACKUP_SOURCE, &source_file) == 0) {
					SS_Free(source_file.data);

					if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
						if (SS_CopyFile(SS_BACKUP_SOURCE, source_filename) != S_SS_SUCCESS) {
							strerror_r(errno, buf, sizeof(buf));
							LOGE("copy of backup to \"%s\" failed: %s\n", source_filename, buf);
							SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
							return E_SS_FAILURE;
						}
						LOGL(LOG_SSENGINE,
							 "Patch Can be applied from using backup file as source\n");
					} else {
						SS_SetUpgradeState(E_SS_FSSRCCURRUPTED);
						return E_SS_FAILURE;
					}
				} else {
					SS_SetUpgradeState(E_SS_FSSRCCURRUPTED);
					return E_SS_FAILURE;
				}
			}
		} else {
			LOGL(LOG_SSENGINE, "Source was deleted!!\n");
			if (SS_LoadFile(target_filename, &target_file) == 0) {
				SS_Free(target_file.data);

				if (memcmp(target_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
					LOGL(LOG_SSENGINE, "Patch Already applied\n");
					return S_SS_SUCCESS;
				}
			}

			LOGL(LOG_SSENGINE, "Try loading from backup source\n");
			if (SS_LoadFile(SS_BACKUP_SOURCE, &source_file) == 0) {
				SS_Free(source_file.data);

				if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
					use_backup = 1;
					LOGL(LOG_SSENGINE, "Patch Can be applied from using backup file as source\n");
				} else {
					SS_SetUpgradeState(E_SS_FSSRCCURRUPTED);
					return E_SS_FAILURE;
				}
			} else {
				LOGE(" SS_LoadFile from backup source failed!!\n");
				SS_SetUpgradeState(E_SS_FSSRCCURRUPTED);
				return E_SS_FAILURE;
			}
		}
	}
	//Now proceed wit patch application since patch can be applied
	do {
		int enough_space = 0;
		size_t free_space = 0;
		char *tok;

		if (retry > 0) {
			if (use_backup) {
				tok = strrchr(source_filename, '/');
				*tok = '\0';
			}
			SS_GetAvailableFreeSpace(source_filename, &free_space);
			enough_space = (free_space > (256 << 10)) &&		// 256k (two-block) minimum
				(free_space > (patch_data_size * 3 / 2));	   // 50% margin of error
			if (use_backup)
				*tok = '/';
		}

		if (!use_backup) {
#ifndef ENHANCED_BSDIFF
			backupsrc = SS_BackupSource(source_filename);
			if (backupsrc != 0) {
				LOGE("failed to Backup source File:[%s] \n", source_filename);
				SS_SetUpgradeState(E_SS_FSSRCBACKUPFAILED);
				return E_SS_FAILURE;
			}
#endif
		}
		if (!enough_space) {
			LOGL(LOG_SSENGINE, "For %s: free space %ld bytes; enough %d\n", source_filename, (long)free_space,
				 enough_space);
			retry = 0;
			use_backup = 1;
			unlink(source_filename);
		}
		//LOGL(LOG_SSENGINE, "For %s: target %ld bytes; free space %ld bytes; enough %d\n",
		// source_filename, (long)patch_data_size, (long)free_space, enough_space);
		//LOGL(LOG_SSENGINE,"Generate Target Space availabitiy [%d]\n", enough_space);

		SinkFn sink = NULL;
		void *token = NULL;
		output = -1;
		outname = NULL;

		{
			// We write the decoded output to "<tgt-file>.patch".
			//allocate some extra space to allow for concatinating ".patch" with the name
			outname = (char *)SS_Malloc(strlen(target_filename) + 10);
			if (outname == NULL) {
				LOGE("SS_Malloc failed!!\n");
				return E_SS_FAILURE;
			}
			snprintf(outname, strlen(target_filename) + 10,
					"%s.patch", target_filename);

			output = open(outname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
			if (output < 0) {
				if (errno == 2) {
					char *dir_path = strrchr(outname, '/');
					*dir_path = '\0';
					// need to create directory as the target may be different from source
					LOGL(LOG_SSENGINE, "need to create directory [%s]\n", outname);
					create_dir(outname, 0755);
					*dir_path = '/';
					output = open(outname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
					if (output < 0) {
						strerror_r(errno, buf, sizeof(buf));
						LOGE("failed to open output file %s: %s\n", outname, buf);
						SS_Free(outname);
						return E_SS_FAILURE;
					}
				}
			}
			sink = ss_fileSink;
			token = &output;
		}
		sha1_init(&ctx1);
		if (use_backup)
			result = SS_ApplyBsdiff(SS_BACKUP_SOURCE, outname, SS_PATCHFILE_SOURCE, sink, token, &ctx1);
		else
			result = SS_ApplyBsdiff((char *)source_filename, outname, SS_PATCHFILE_SOURCE, sink, token, &ctx1);
		if (output >= 0) {
			fsync(output);
			close(output);
		}

		if (result != S_SS_SUCCESS) {
			if (retry == 0) {
				LOGE("applying patch failed result : [%d]\n", result);
				SS_Free(outname);//wgid: 20739
				SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
				return E_SS_FAILURE;
			} else {
				LOGE("applying patch failed; retrying\n");
				SS_Free(outname);//wgid: 20739
			}
		} else {
			// succeeded; no need to retry
			break;
		}
	} while (retry-- > 0);

	const uint8_t current_target_sha1[SHA_DIGEST_SIZE] = { 0, };
	sha1_final(&ctx1, (uint32_t *) & current_target_sha1);
	if (memcmp(current_target_sha1, target_sha1, SHA_DIGEST_SIZE) != 0) {
		LOGE("patch did not produce expected sha1\n");
		SS_SetUpgradeState(E_SS_FSSHA_MISMATCH);
		if (outname != NULL)
			SS_Free(outname);
		return E_SS_FAILURE;
	}

	// Finally, rename the .patch file to replace the target file.
#ifdef ENHANCED_BSDIFF
	if (SS_rename1(outname, target_filename) != 0) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("rename of .patch to \"%s\" failed: %s\n", target_filename, buf);
		SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
		if (outname != NULL)
			SS_Free(outname);
		return E_SS_FAILURE;
	}
#else
	if (rename(outname, target_filename) != 0) {
		strerror_r(errno, buf, sizeof(buf));
		LOGE("rename of .patch to \"%s\" failed: %s\n", target_filename, buf);
		SS_SetUpgradeState(E_SS_FSUPDATEFAILED);
		if (outname != NULL)
			SS_Free(outname);
		return E_SS_FAILURE;
	}
	//remove source file if target is not same
	if (strcmp(source_filename, target_filename) != 0)
		unlink(source_filename);
	SS_BackupSourceClear();
#endif
	SS_PatchSourceClear();
	SS_Free(outname);

	return result;
}

/*!
 *********************************************************************************
 *					 SS_UpdateDeltaKernel
 *********************************************************************************
 *
 * @brief
 *	This is used to apply patch for kernel delta during delta Image upgrade
 *
 *
 *	@param
 *
 *	@return				0 - in case of success
 *						1 - in case of error during patch application
 *
 *********************************************************************************
 */

int SS_UpdateDeltaKernel(ua_dataSS_t * ua_dataSS, int (*write_to_blkdev) (char *, int, int, char *))
{
	uint8_t target_sha1[SHA_DIGEST_SIZE];
	uint8_t source_sha1[SHA_DIGEST_SIZE];
	FileInfo source_file;
	int result = S_SS_SUCCESS;
	int blk_cnt, read_count = 0;
	int blk_start = 0;
	int backupsrc = -1;
	int use_backup_img = -1;
	FILE *fp = NULL, *wp = NULL, *kp = NULL;
	int i = 0, j = 0, file_len = 0;
	char *magic = NULL, *file_name = NULL, *tok_buf = NULL, *buf = NULL, a = '0';
	char *saveptr;
	char err_buf[256];
	char cmd[1024] = { 0, };
	char source_filename[MAX_FILE_PATH] = { 0, };
	char part_filename[MAX_FILE_PATH] = { 0, };
	int file_num = 0;
	SinkFn sink = NULL;
	void *tok = NULL;
	int output = -1;
	char *outname = NULL;
	//Kernel Parts are created on unpacking kernel which is then used to apply delta
	char *kernel_parts[] = { "decompression_code",
		"piggy.gz",
		"padding_piggy",
		"piggy_trailer"
	};

	if (ParseSha1(ua_dataSS->update_cfg->target_sha1, target_sha1) != 0) {
		LOGE("failed to parse tgt-sha1 \"%s\"\n", ua_dataSS->update_cfg->target_sha1);
		return E_SS_FAILURE;
	}

	source_file.size = ua_dataSS->update_cfg->soure_img_size;
	source_file.data = NULL;
	if (0 == gvalid_session) {

		if (ParseSha1(ua_dataSS->update_cfg->soure_sha1, source_sha1) != 0) {
			LOGE("failed to parse Src-sha1 \"%s\"\n", ua_dataSS->update_cfg->soure_sha1);
			return E_SS_FAILURE;
		}

		if (SS_LoadPartition(ua_dataSS->parti_info->ua_blk_name, &source_file) == 0) {
			if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Patch Can be applied\n");

			} else if (memcmp(source_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Patch Already applied\n");
				SS_Free(source_file.data);
				return S_SS_SUCCESS;
			} else {
				SS_Free(source_file.data);
				source_file.data = NULL;
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Source was currupted, Try loading from backup source\n");
				if (SS_LoadPartition(SS_BACKUP_SOURCE, &source_file) == 0) {
					if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
						use_backup_img = 1;

						LOGL(LOG_SSENGINE,
							 "SS_UpdateDeltaIMG - Patch Can be applied from using backup file as source\n");
					} else {
						SS_SetUpgradeState(E_SS_IMGSRCCURRUPTED);
						SS_Free(source_file.data);
						return E_SS_FAILURE;
					}
				}
			}
		}
	} else {					//in case of kernel delta need to copy kernel data  from blk to buffer
		if (SS_LoadPartition(ua_dataSS->parti_info->ua_blk_name, &source_file) != 0) {
			SS_SetUpgradeState(E_SS_IMGSRCCURRUPTED);
			LOGE("Fatal Error : Kernel block is corrupted\n");
			return E_SS_FAILURE;
		}
	}
	if (use_backup_img == -1) {
		backupsrc = SS_BackupSource(ua_dataSS->parti_info->ua_blk_name);
		if (backupsrc != 0) {
			LOGE("failed to Backup source File:[%s] \n", ua_dataSS->parti_info->ua_blk_name);
			SS_Free(source_file.data);
			SS_SetUpgradeState(E_SS_IMGSRCBACKUPFAILED);
			return E_SS_FAILURE;
		}
	}
	//Cleanup workspace and copy helper executables to it before proceeding
	SS_DeleteFolder(SS_KERNEL_WORKSPACE);
	create_dir(SS_KERNEL_WORKSPACE, 0755);
	SS_CopyFile(SS_GZIP_SOURCE, SS_GZIP_TARGET);
	SS_CopyFile(SS_STAT_SOURCE, SS_STAT_TARGET);
	SS_CopyFile(SS_DD_SOURCE, SS_DD_TARGET);

	if (tar_get_item_size(ua_dataSS->update_data->ua_delta_path, SS_KERNEL_UNPACK_SCRIPT) > 0)
		if (tar_extract_file(ua_dataSS->update_data->ua_delta_path, SS_KERNEL_UNPACK_SCRIPT, SS_KERN_UNPK_SCRIPT_PATH) > 0)
			LOGL(LOG_SSENGINE, "Extracted %s successfully\n", SS_KERNEL_UNPACK_SCRIPT);
		else {
			LOGE("Error in fn tar_extract_file for item %s", SS_KERNEL_UNPACK_SCRIPT);
			SS_SetUpgradeState(E_SS_DELTA_IS_CORRUPT);
			result = E_SS_FAILURE;
			goto Cleanup;
	} else {
		LOGE("Error size is not positive for item %s", SS_KERNEL_UNPACK_SCRIPT);
		SS_SetUpgradeState(E_SS_DELTA_IS_CORRUPT);
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	//Now write the kernel data to the workplace and start applying patch
	snprintf(source_filename, sizeof(source_filename) - 1, "%s/%s", SS_KERNEL_WORKSPACE, SS_KERNEL_NAME);
	fp = fopen(source_filename, "w");
	if (!fp) {
		LOGE("file open error [%s] code [%d]\n", source_filename, errno);
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	//write source kernel data to workspace
	read_count = fwrite(source_file.data, 1, source_file.size, fp);
	if (read_count != source_file.size) {
		LOGE("file write error read_count = %d for [%s]\n", read_count, source_filename);
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	fclose(fp);
	fp = NULL;//wgid: 59313

	//Unpack source kernel
	int offset = getOffset(source_filename);
	if (offset < 0) {
		LOGE("Failed to get offset\n");
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	snprintf(cmd, sizeof(cmd) - 1, "%s -u %s %s %d", SS_KERN_UNPK_SCRIPT_PATH, SS_KERNEL_WORKSPACE, SS_KERNEL_NAME,
			 offset);
	int ret = _system_cmd_wait(cmd);
	LOGL(LOG_SSENGINE, "ret for %s is %d\n", cmd, ret);

	//open delta file, extract kernel delta parts and apply patch to previously unpacked kernel
	fp = fopen(SS_PATCHFILE_SOURCE, "r");
	if (fp == NULL) {
		LOGE("Failed to open kernel delta patch\n");
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	//read kernel delta header for  delta names and size
	buf = SS_Malloc(SS_KERNEL_DELTA_HEADER);
	if (!buf) {//wgid: 13099
		LOGE("Failed to allocate memory\n");
		result = E_SS_MALLOC_ERROR;
		goto Cleanup;
	}
	ssize_t bytes_read = fread(buf, 1, SS_KERNEL_DELTA_HEADER, fp);
	if (bytes_read != SS_KERNEL_DELTA_HEADER)
		LOGL(LOG_SSENGINE, "short read of \"%s\" (%ld bytes of %ld)\n", SS_PATCHFILE_SOURCE, (long)bytes_read, (long)SS_KERNEL_DELTA_HEADER);
	magic = strtok_r(buf, ":", &saveptr);
	LOGL(LOG_SSENGINE, "magic: %s\n", magic);
	tok_buf = strtok_r(NULL, ":", &saveptr);
	if (tok_buf)
		file_num = atoi(tok_buf);

	//adjust offset to start of data section before proceeding
	fseek(fp, SS_KERNEL_DELTA_HEADER, SEEK_SET);

	while (file_num-- > 0) {
		file_name = strtok_r(NULL, ":", &saveptr);
		tok_buf = strtok_r(NULL, ":", &saveptr);
		if (tok_buf)
			file_len = atoi(tok_buf);
		j = file_len;
		snprintf(source_filename, sizeof(source_filename) - 1, "%s/%s_unpacked/%s", SS_KERNEL_WORKSPACE, SS_KERNEL_NAME,
				 file_name);
		snprintf(part_filename, sizeof(part_filename) - 1, "%s/%s", SS_KERNEL_WORKSPACE, file_name);
		wp = fopen(part_filename, "w");
		if (wp == NULL) {
			LOGE("Failed to open %s\n", part_filename);
			result = E_SS_FAILURE;
			goto Cleanup;
		}
		while (j-- > 0) {
			a = fgetc(fp);
			fputc(a, wp);
		}
		fclose(wp);

		//apply bspatch to the unpacked kernel parts

		{
			// We write the decoded output to "<tgt-file>.patch".
			//allocate some extra space to allow for concatinating ".patch" with the name
			outname = (char *)SS_Malloc(strlen(source_filename) + 10);
			if (outname == NULL) {
				SS_SetUpgradeState(E_SS_MALLOC_ERROR);
				result = E_SS_FAILURE;
				goto Cleanup;
			}
			snprintf(outname, strlen(source_filename) + 10, "%s.patch",
					source_filename);

			output = open(outname, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
			if (output < 0) {
				strerror_r(errno, err_buf, sizeof(err_buf));
				LOGE("failed to open output file %s: %s\n", outname, err_buf);
				result = E_SS_FAILURE;
				goto Cleanup;
			}
			sink = ss_fileSink;
			tok = &output;
		}
		result = SS_ApplyBsdiff(source_filename, outname, part_filename, sink, tok, NULL);
		LOGL(LOG_SSENGINE, "GenerateTarget Output is %d and result is %d\n", output, result);
		if (output >= 0) {
			fsync(output);
			close(output);
		}

		if (result != S_SS_SUCCESS) {
			LOGE("applying patch failed %s\n", source_filename);
			if (outname != NULL)
				unlink(outname);
			goto Cleanup;
		}
		result = rename(outname, source_filename);
		if (result != S_SS_SUCCESS) {
			LOGE("fatal error %s\n", source_filename);
			goto Cleanup;
		}
		if (file_name && strcmp(file_name, "piggy") == 0) {
			snprintf(cmd, sizeof(cmd) - 1,
					 "%s/gzip -n -9 -c %s/%s/%s > %s/%s/%s.gz",
					 SS_KERNEL_WORKSPACE, SS_KERNEL_WORKSPACE, SS_KERNEL_UNPACK_DIR, file_name,
					 SS_KERNEL_WORKSPACE, SS_KERNEL_UNPACK_DIR, file_name);
			result = _system_cmd_wait(cmd);
			LOGL(LOG_SSENGINE, "ret for %s = %d\n", cmd, result);
			unlink(source_filename);
		}
		unlink(part_filename);
		SS_Free(outname);//wgid :144116
		outname = NULL;
	}
	//open new kernel file and append kernel parts to it in
	snprintf(source_filename, sizeof(source_filename) - 1, "%s/%s", SS_KERNEL_WORKSPACE, SS_KERNEL_TARGET_NAME);
	kp = fopen(source_filename, "w");
	if (kp == NULL) {
		LOGE("Failed to open %s\n", source_filename);
		result = E_SS_FAILURE;
		goto Cleanup;
	}
	for (i = 0; i < 4; i++) {
		snprintf(part_filename, sizeof(part_filename) - 1, "%s/%s/%s", SS_KERNEL_WORKSPACE, SS_KERNEL_UNPACK_DIR,
				 kernel_parts[i]);
		wp = fopen(part_filename, "r");
		if (wp == NULL) {
			LOGE("Failed to open %s\n", part_filename);
			result = E_SS_FAILURE;
			fclose(kp);
			goto Cleanup;
		}
		fseek(wp, SEEK_SET, SEEK_END);
		j = ftell(wp);
		fseek(wp, SEEK_SET, SEEK_SET);
		while (j-- > 0) {
			a = fgetc(wp);
			if (a != EOF)
				fputc(a, kp);
			else
				break;
		}
		fclose(wp);
		unlink(part_filename);
	}
	fclose(fp);
	fp = NULL;
	fclose(kp);

	SS_Free(source_file.data);
	source_file.data = NULL;
	if (SS_LoadFile(source_filename, &source_file) == 0)
		result = memcmp(source_file.sha1, target_sha1, SHA_DIGEST_SIZE);
	if (result != S_SS_SUCCESS) {
		LOGE("patch did not produce expected sha1 \n");
		SS_SetUpgradeState(E_SS_IMGSHA_MISMATCH);
		goto Cleanup;
	}
	//Considering EMMC partition by deafult

	blk_cnt = ((ua_dataSS->update_cfg->target_img_size - 1) / SECTOR_SIZE) + 1;
	result = write_to_blkdev((char *)ua_dataSS->parti_info->ua_blk_name, blk_start, blk_cnt, (char *)source_file.data);
	if (result != S_SS_SUCCESS) {
		LOGE("write of patched data to %s failed\n", ua_dataSS->parti_info->ua_blk_name);	   // All returns should go to CLEAN UP.
		SS_SetUpgradeState(E_SS_IMGFLASHWRITEFAIL);
		goto Cleanup;
	}

 Cleanup:
	SS_BackupSourceClear();
	SS_PatchSourceClear();
	SS_DeleteFile(SS_KERN_UNPK_SCRIPT_PATH);
	SS_DeleteFolder(SS_KERNEL_WORKSPACE);
	SS_Free(source_file.data);
	SS_Free(buf);
	SS_Free(outname);//wgid: 20740
	if (result == S_SS_SUCCESS)
		LOGL(LOG_SSENGINE, "************* SS_UpdateDeltaKernel  SUCCESS *****************\n");
	else{
		LOGL(LOG_SSENGINE, "************* SS_UpdateDeltaKernel  FAILED *****************\n");
		if (fp)
			fclose(fp);//wgid:14711
	}
	return result;

}

/*!
 *********************************************************************************
 *					 SS_UpdateDeltaIMG
 *********************************************************************************
 *
 * @brief
 *	This is used to apply patch for an image during delta Image upgrade
 *
 *
 *	@param
 *
 *	@return				0 - in case of success
 *						1 - in case of error during patch application
 *
 *********************************************************************************
 */

int SS_UpdateDeltaIMG(ua_dataSS_t * ua_dataSS, int (*write_to_blkdev) (char *, int, int, char *))
{
	uint8_t target_sha1[SHA_DIGEST_SIZE];
	uint8_t source_sha1[SHA_DIGEST_SIZE];
	const uint8_t current_target_sha1[SHA_DIGEST_SIZE];
	FileInfo source_file;
	sha1_ctx_t ctx1;
	MemorySinkInfo msi;
	int result = S_SS_SUCCESS;
	int blk_cnt;
	int blk_start = 0;
	int backupsrc = -1;
	int use_backup_img = -1;
	int fd = -1;
	char buf[256];
	if (ParseSha1(ua_dataSS->update_cfg->target_sha1, target_sha1) != 0) {
		LOGE("failed to parse tgt-sha1 \"%s\"\n", ua_dataSS->update_cfg->target_sha1);
		return E_SS_FAILURE;
	}

	source_file.size = ua_dataSS->update_cfg->soure_img_size;
	source_file.data = NULL;
	if (0 == gvalid_session) {

		if (ParseSha1(ua_dataSS->update_cfg->soure_sha1, source_sha1) != 0) {
			LOGE("failed to parse Src-sha1 \"%s\"\n", ua_dataSS->update_cfg->soure_sha1);
			return E_SS_FAILURE;
		}

		if (SS_LoadPartition(ua_dataSS->parti_info->ua_blk_name, &source_file) == 0) {
			if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Patch Can be applied\n");
				SS_Free(source_file.data);
			} else if (memcmp(source_file.sha1, target_sha1, SHA_DIGEST_SIZE) == 0) {
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Patch Already applied\n");
				SS_Free(source_file.data);
				return S_SS_SUCCESS;
			} else {
				SS_Free(source_file.data);
				source_file.data = NULL;
				LOGL(LOG_SSENGINE, "SS_UpdateDeltaIMG - Source was currupted, Try loading from backup source\n");
				if (SS_LoadPartition(SS_BACKUP_SOURCE, &source_file) == 0) {
					if (memcmp(source_file.sha1, source_sha1, SHA_DIGEST_SIZE) == 0) {
						use_backup_img = 1;
						SS_Free(source_file.data);
						LOGL(LOG_SSENGINE,
							 "SS_UpdateDeltaIMG - Patch Can be applied from using backup file as source\n");
					} else {
						SS_SetUpgradeState(E_SS_IMGSRCCURRUPTED);
						SS_Free(source_file.data);
						return E_SS_FAILURE;
					}
				}
			}
		}
	}
	if (use_backup_img == -1) {
		backupsrc = SS_BackupSource(ua_dataSS->parti_info->ua_blk_name);
		if (backupsrc != 0) {
			LOGE("failed to Backup source File:[%s] \n", ua_dataSS->parti_info->ua_blk_name);
			SS_SetUpgradeState(E_SS_IMGSRCBACKUPFAILED);
			return E_SS_FAILURE;
		}
	}
	SinkFn sink = NULL;
	void *token = NULL;

	blk_cnt = ((ua_dataSS->update_cfg->target_img_size - 1) / SECTOR_SIZE) + 1;

	msi.buffer = SS_Malloc(blk_cnt * SECTOR_SIZE);
	if (msi.buffer == NULL) {
		LOGE("failed to alloc %ld bytes for output\n", (long)ua_dataSS->update_cfg->target_img_size);
		SS_SetUpgradeState(E_SS_MALLOC_ERROR);
		return E_SS_FAILURE;
	}
	msi.pos = 0;
	msi.size = ua_dataSS->update_cfg->target_img_size;
	sink = ss_memorySink;
	token = &msi;

	sha1_init(&ctx1);
	//if souce was corrupted, use backup to apply diff
	if (use_backup_img == -1)
		result =
			SS_ApplyBsdiff((char *)ua_dataSS->parti_info->ua_blk_name, NULL, SS_PATCHFILE_SOURCE, sink, token, &ctx1);
	else
		result = SS_ApplyBsdiff(SS_BACKUP_SOURCE, NULL, SS_PATCHFILE_SOURCE, sink, token, &ctx1);
	if (result != S_SS_SUCCESS) {
		LOGE("failed to SS_ApplyBsdiff\n");
		SS_SetUpgradeState(E_SS_IMGRECOVERYWRITEFAILED);
		goto Cleanup;
	}

	sha1_final(&ctx1, (uint32_t *) & current_target_sha1);
	result = memcmp(current_target_sha1, target_sha1, SHA_DIGEST_SIZE);
	if (result != S_SS_SUCCESS) {
		LOGE("patch did not produce expected sha1 \n");
		SS_SetUpgradeState(E_SS_IMGSHA_MISMATCH);
		goto Cleanup;
	}
	//Considering EMMC partition by deafult

	if (ua_dataSS->update_cfg->update_type == DELTA_IMG) {
		blk_cnt = ((ua_dataSS->update_cfg->target_img_size - 1) / SECTOR_SIZE) + 1;
		result = write_to_blkdev((char *)ua_dataSS->parti_info->ua_blk_name, blk_start, blk_cnt, (char *)msi.buffer);
		if (result != S_SS_SUCCESS) {
			LOGE("write of patched data to %s failed\n", ua_dataSS->parti_info->ua_blk_name);   // All returns should go to CLEAN UP.
			SS_SetUpgradeState(E_SS_IMGFLASHWRITEFAIL);
			goto Cleanup;
		}
	} else if (ua_dataSS->update_cfg->update_type == EXTRA && ua_dataSS->update_data->ua_temp_path) {
		fd = open(ua_dataSS->update_data->ua_temp_path, O_WRONLY | O_CREAT | O_TRUNC, S_IRWXU);
		if (fd < 0) {
			strerror_r(errno, buf, sizeof(buf));
			LOGE("failed to open %s for write: %s\n", ua_dataSS->update_data->ua_temp_path, buf);
			SS_SetUpgradeState(E_SS_IMGRECOVERYWRITEFAILED);
			result = E_SS_FAILURE;
			goto Cleanup;
		}
		result = SS_WriteFile(fd, 0, msi.buffer, msi.size);
		if (result != S_SS_SUCCESS) {
			LOGE("failed to write\n");
			SS_SetUpgradeState(E_SS_IMGRECOVERYWRITEFAILED);
			goto Cleanup;
		}
		fsync(fd);
	} else {
		SS_SetUpgradeState(E_SS_IMGUPDATEFAILED);
		result = E_SS_FAILURE;
		LOGE("failed to apply patch - Invalid Update type params \n");
	}

Cleanup:
	SS_BackupSourceClear();
	SS_PatchSourceClear();
	if (msi.buffer)
		SS_Free(msi.buffer);
	if (fd >= 0)
		close(fd);
	if (result == S_SS_SUCCESS)
		LOGL(LOG_SSENGINE, "************* SS_UpdateDeltaIMG  SUCCESS *****************\n");
	return result;

}
