/*
 * libtota
 *
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *	 http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include "SS_Engine_Update.h"
#include "SS_Engine_Errors.h"
#include "SS_FSUpdate.h"
#include "fota_common.h"
#include "fota_tar.h"
#include "ua_types.h"

/* tar Header Block, from POSIX 1003.1-1990. for reference */
#if 0
		/* POSIX header.  */

struct posix_header {		   /* byte offset */
	char name[100];			 /*   0 */
	char mode[8];			   /* 100 */
	char uid[8];				/* 108 */
	char gid[8];				/* 116 */
	char size[12];			  /* 124 */
	char mtime[12];			 /* 136 */
	char chksum[8];			 /* 148 */
	char typeflag;			  /* 156 */
	char linkname[100];		 /* 157 */
	char magic[6];			  /* 257 */
	char version[2];			/* 263 */
	char uname[32];			 /* 265 */
	char gname[32];			 /* 297 */
	char devmajor[8];		   /* 329 */
	char devminor[8];		   /* 337 */
	char prefix[155];		   /* 345 */
	/* 500 */
};
#endif

#define	MAX_ITEM_SIZE				0x7FFFFFFF
#define	TAR_ITEM_SIZE_POSITION	124
#define	TAR_SIZE_OF_ITEM_SIZE		8
#define	TAR_SIZE_OF_HEADER		12
#define	TAR_BLOCK_SIZE				512
#define	TAR_ITEM_NAME_SIZE		100
#define	TAR_LONG_NAME_SIZE		256
#define	TAR_ITEM_TYPE_FLAG_POS	156

/*** The byte that indicates whether the prefix is present or not */
#define PREFIX_INDICATOR_BYTE 345
#define PREFIX_LEN 155

/** the rest heavily based on (ie mostly) untgz.c from zlib **/

/* Values used in typeflag field.  */

#define REGTYPE  '0'			/* regular file */
#define AREGTYPE '\0'		   /* regular file */
#define LNKTYPE  '1'			/* link */
#define SYMTYPE  '2'			/* reserved */
#define CHRTYPE  '3'			/* character special */
#define BLKTYPE  '4'			/* block special */
#define DIRTYPE  '5'			/* directory */
#define FIFOTYPE '6'			/* FIFO special */
#define CONTTYPE '7'			/* reserved, for compatibility with gnu tar,
								   treat as regular file, where it represents
								   a regular file, but saved contiguously on disk */

/* GNU tar extensions */

#define GNUTYPE_DUMPDIR  'D'	/* file names from dumped directory */
#define GNUTYPE_LONGLINK 'K'	/* long link name */
#define GNUTYPE_LONGNAME 'L'	/* long file name */
#define GNUTYPE_MULTIVOL 'M'	/* continuation of file from another volume */
#define GNUTYPE_NAMES	'N'	/* file name that does not fit into main hdr */
#define GNUTYPE_SPARSE   'S'	/* sparse file */
#define GNUTYPE_VOLHDR   'V'	/* tape/volume header */

extern void *SS_Malloc(SS_UINT32 size);

int gTarFd = -1;				// Currenlty this logic supports only one tar file

/* Parse an octal number, ignoring leading and trailing nonsense. */
static int parseoct(const char *p, size_t n)
{
	int i = 0;

	while (*p < '0' || *p > '7') {
		++p;
		--n;
	}
	while (*p >= '0' && *p <= '7' && n > 0) {
		i *= 8;
		i += *p - '0';
		++p;
		--n;
	}
	return (i);
}

/* Verify the tar checksum. */
static int verify_checksum(const char *p)
{
	int n, u = 0;
	for (n = 0; n < 512; ++n) {
		if (n < 148 || n > 155)
			/* Standard tar checksum adds unsigned bytes. */
			u += ((unsigned char *)p)[n];
		else
			u += 0x20;

	}
	return (u == parseoct(p + 148, 8));
}

static int is_end_of_archive(const char *p)
{
	int n;
	for (n = 511; n >= 0; --n)
		if (p[n] != '\0')
			return (0);
	return (1);
}

void create_dir(char *pathname, int mode)
{
	char *p;
	int r;

	/* Strip trailing '/' */
	if (pathname[strlen(pathname) - 1] == '/')
		pathname[strlen(pathname) - 1] = '\0';

	/* Try creating the directory. */
	r = mkdir(pathname, mode);

	if (r != 0) {
		/* On failure, try creating parent directory. */
		p = strrchr(pathname, '/');
		if (p != NULL) {
			*p = '\0';
			create_dir(pathname, 0755);
			*p = '/';
			r = mkdir(pathname, mode);
		}
	}
	if (r != 0) {
		if (r != EEXIST && r != -1)
			LOG("Could not create directory [%s] Error[%d]\n", pathname, r);
	}
}

/* Create a file, including parent directory as necessary. */
static FILE *create_file(char *pathname)
{
	FILE *f;
	f = fopen(pathname, "w+");
	if (f == NULL) {
		/* Try creating parent dir and then creating file. */
		char *p = strrchr(pathname, '/');
		if (p != NULL) {
			*p = '\0';
			create_dir(pathname, 0755);
			*p = '/';
			f = fopen(pathname, "w+");
		}
	}
	return (f);
}

/*-----------------------------------------------------------------------------
  tar_get_item_offset
 ----------------------------------------------------------------------------*/
int tar_get_item_offset(char *tar, char *item)
{
	int ret = -1;
	int fd = -1;
	char header[TAR_BLOCK_SIZE] = { 0, };
	char name[TAR_ITEM_NAME_SIZE + 1] = { 0, };
	char size_oct[TAR_SIZE_OF_HEADER] = { 0, };
	unsigned long size_dec = 0;
	int blknum = 0;
	off_t pos = 0;
	off_t tar_len = 0;
	ssize_t rdcnt = 0;

	//check if gTarFd was opened by tar_open during SS_FSUpdateFile then use it
	if (gTarFd > 0)
		fd = gTarFd;
	if (fd < 0) {
		fd = open(tar, O_RDONLY);
		if (fd < 0) {
			LOGE("can't open file(%s).\n", tar);
			return -1;
		}
	}

	tar_len = lseek(fd, 0, SEEK_END);
	if (tar_len < 0) {
		LOGL(LOG_SSENGINE, "can't read tar_len (%s).\n", tar);
		goto Cleanup;
	}
	pos = lseek(fd, 0, SEEK_SET);
	if (pos < 0) {
		LOGL(LOG_SSENGINE, "can't read pos (%s).\n", tar);
		goto Cleanup;
	}
	while (pos < tar_len) {
		/* read file header */
		rdcnt = read(fd, header, sizeof(header));
		if (rdcnt <= 0) {
			LOG("read failed. (rdcnt=%d)\n", rdcnt);
			break;
		}

		/* get file name and file size */
		memcpy(name, header, sizeof(name) - 1);
		memcpy(size_oct, header + TAR_ITEM_SIZE_POSITION, sizeof(size_oct));
		size_dec = strtoul(size_oct, NULL, TAR_SIZE_OF_ITEM_SIZE);
		if (size_dec > MAX_ITEM_SIZE) {
			LOG("size too big. (size_dec=0x%08X)\n", size_dec);
			break;
		}

		/* check if the file is what we are looking for */
		if (strncmp(name, item, TAR_ITEM_NAME_SIZE) == 0) {
			ret = (int)lseek(fd, 0, SEEK_CUR);
			break;
		}

		/* move file pointer to next file header */
		blknum = size_dec / TAR_BLOCK_SIZE;
		if (size_dec % TAR_BLOCK_SIZE)
			blknum++;

		pos = lseek(fd, (off_t) (blknum * TAR_BLOCK_SIZE), SEEK_CUR);
		if (pos < 0) {
			LOGE("can't read next block (%s).\n", tar);
			close(fd);
			return -1;
		}
	}

 Cleanup:
	if (gTarFd < 0)
		close(fd);

	return ret;
}

/*-----------------------------------------------------------------------------
  tar_get_item_size
 ----------------------------------------------------------------------------*/
int tar_get_item_size(char *tar, char *item)
{
	int ret = -1;
	int fd = -1;
	char header[TAR_BLOCK_SIZE] = { 0, };
	char uExtendedName[MAX_FILE_PATH] = { 0, };
	char size_oct[TAR_SIZE_OF_HEADER] = { 0, };
	unsigned long size_dec = 0;
	int blknum = 0;
	off_t pos = 0;
	off_t tar_len = 0;
	ssize_t rdcnt = 0;
	LOGL(LOG_SSENGINE, "Tar file Looking for (%s)\n", item);
	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOG("can't open file(%s).\n", tar);
		return -1;
	}

	tar_len = lseek(fd, 0, SEEK_END);
	if (tar_len < 0) {
		LOGL(LOG_SSENGINE, "can't read tar_len (%s).\n", tar);
		goto Cleanup;
	}
	pos = lseek(fd, 0, SEEK_SET);
	if (pos < 0) {
		LOGL(LOG_SSENGINE, "can't read pos (%s).\n", tar);
		goto Cleanup;
	}

	while (pos < tar_len) {
		/* read file header */
		rdcnt = read(fd, header, sizeof(header));
		if (rdcnt <= 0) {
			LOG("read failed. (rdcnt=%d)\n", rdcnt);
			break;
		}

		/* get file name and file size */
		if (header[TAR_ITEM_TYPE_FLAG_POS] == GNUTYPE_LONGNAME || header[TAR_ITEM_TYPE_FLAG_POS] == GNUTYPE_LONGLINK) {
			//rdcnt = read(fd, header, sizeof(header));
			memset(uExtendedName, 0, sizeof(uExtendedName));
			rdcnt = read(fd, uExtendedName, sizeof(uExtendedName));
			if (rdcnt <= 0) {
				LOG("read failed. (rdcnt=%d)\n", rdcnt);
				break;
			}
			rdcnt = read(fd, header, sizeof(header));
			if (rdcnt <= 0) {
				LOG("read failed. (rdcnt=%d)\n", rdcnt);
				break;
			}
		} else {
			memset(uExtendedName, 0, sizeof(uExtendedName));
			memcpy(uExtendedName, header, TAR_ITEM_NAME_SIZE);
		}
		//memcpy(name, header, sizeof(name));
		memcpy(size_oct, header + TAR_ITEM_SIZE_POSITION, sizeof(size_oct));
		size_dec = strtoul(size_oct, NULL, TAR_SIZE_OF_ITEM_SIZE);
		if (size_dec > MAX_ITEM_SIZE) {
			LOG("size too big. (size_dec=0x%08X)\n", (unsigned int)size_dec);
			break;
		}

		/* check if the file is what we are looking for */
		if (strcmp(uExtendedName, item) == 0) {
			ret = (int)size_dec;
			if ((ret == 0) && (header[TAR_ITEM_TYPE_FLAG_POS] == DIRTYPE))
				ret = tar_get_folder_size(tar, item);
			break;
		}
		/* move file pointer to next file header */
		//LOGL(LOG_SSENGINE,"Item in Tar (%s)\n", uExtendedName);
		blknum = size_dec / TAR_BLOCK_SIZE;
		if (size_dec % TAR_BLOCK_SIZE)
			blknum++;

		pos = lseek(fd, (off_t) (blknum * TAR_BLOCK_SIZE), SEEK_CUR);
		if (pos < 0) {
			LOGL(LOG_SSENGINE, "can't read next block (%s).\n", tar);
			close(fd);
			return -1;
		}
	}

 Cleanup:
	close(fd);

	return ret;
}

/*-----------------------------------------------------------------------------
  tar_get_item_tye. (Dir/file/link etc)
 ----------------------------------------------------------------------------*/

char tar_get_item_type(char *tar, char *item)
{
	char ret = '0';
	int fd = -1;
	char header[TAR_BLOCK_SIZE] = { 0, };
	char name[TAR_ITEM_NAME_SIZE + 1] = { 0, };
	char size_oct[TAR_SIZE_OF_HEADER] = { 0, };
	unsigned long size_dec = 0;
	int blknum = 0;
	off_t pos = 0;
	off_t tar_len = 0;
	ssize_t rdcnt = 0;
	//LOG("Tar file Looking for (%s)\n", item);
	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOG("can't open file(%s).\n", tar);
		return -1;
	}

	tar_len = lseek(fd, 0, SEEK_END);
	if (tar_len < 0) {
		LOGL(LOG_SSENGINE, "can't read tar_len (%s).\n", tar);
		goto Cleanup;
	}
	pos = lseek(fd, 0, SEEK_SET);
	if (pos < 0) {
		LOGL(LOG_SSENGINE, "can't read pos (%s).\n", tar);
		goto Cleanup;
	}

	while (pos < tar_len) {
		/* read file header */
		rdcnt = read(fd, header, sizeof(header));
		if (rdcnt <= 0) {
			LOG("read failed. (rdcnt=%d)\n", rdcnt);
			ret = -1;
			break;
		}

		/* get file name and file size */
		memcpy(name, header, sizeof(name) - 1);
		memcpy(size_oct, header + TAR_ITEM_SIZE_POSITION, sizeof(size_oct));
		size_dec = strtoul(size_oct, NULL, TAR_SIZE_OF_ITEM_SIZE);
		if (size_dec > MAX_ITEM_SIZE) {
			LOG("size too big. (size_dec=0x%08X)\n", (unsigned int)size_dec);
			ret = -1;
			break;
		}

		/* check if the file is what we are looking for */
		if (strncmp(name, item, TAR_ITEM_NAME_SIZE) == 0) {
			ret = header[TAR_ITEM_TYPE_FLAG_POS];
			break;
		}

		/* move file pointer to next file header */
		blknum = size_dec / TAR_BLOCK_SIZE;
		if (size_dec % TAR_BLOCK_SIZE)
			blknum++;

		pos = lseek(fd, (off_t) (blknum * TAR_BLOCK_SIZE), SEEK_CUR);
		if (pos < 0) {
			LOGL(LOG_SSENGINE, "can't read next block (%s).\n", tar);
			close(fd);
			return -1;
		}
	}

 Cleanup:
	close(fd);

	return ret;
}

/*-----------------------------------------------------------------------------
  tar_get_cfg_data
 ----------------------------------------------------------------------------*/
int tar_get_cfg_data(char *tar, char *item, char *buf, int buflen)
{
	int fd = -1;
	int data_size = -1;
	int data_offset = -1;
	off_t pos = 0;
	ssize_t rdcnt = 0;

	data_size = tar_get_item_size(tar, item);
	if (data_size <= 0)
		return -1;

	if (data_size > buflen)
		data_size = buflen;

	data_offset = tar_get_item_offset(tar, item);
	if (data_offset < 0)
		return -1;

	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOG("can't open file(%s).\n", tar);
		return -1;
	}

	pos = lseek(fd, data_offset, SEEK_SET);
	if (pos < 0) {
		LOG("lseek fail (%s offset %d).\n", tar, data_offset);
		close(fd);
		return -1;
	}

	rdcnt = read(fd, buf, data_size);
	if (rdcnt != (ssize_t) data_size) {
		LOG("read fail(%s from %s).\n", item, tar);
		close(fd);
		return -1;
	}

	close(fd);

	return rdcnt;
}

tar_Data_t *tar_build_cfg_table(char *tar)
{

	int fd = -1;
	int ret = 0;
	char header[TAR_BLOCK_SIZE] = { 0, };
	char uExtendedName[MAX_FILE_PATH] = { 0, };
	char size_oct[TAR_SIZE_OF_HEADER] = { 0, };
	unsigned long size_dec = 0;
	int blknum = 0;
	off_t pos = 0;
	off_t tar_len = 0;
	ssize_t rdcnt = 0;

	int itemSize;
	int itemOffset;
	tar_Data_t *headparam = NULL, *tailparam = NULL, *newnode = NULL;
	tar_Data_t *local_temp = NULL;
	tar_Data_t *local_next = NULL;
	if (!tar) {
		LOGE("Bad param tar\n");
		return NULL;
	}
	//check if gTarFd was opened by tar_open during SS_FSUpdateFile then use it
	if (gTarFd > 0)
		fd = gTarFd;
	if (fd < 0) {
		fd = open(tar, O_RDONLY);
		if (fd < 0) {
			LOG("can't open file(%s).\n", tar);
			return NULL;
		}
	}
	tar_len = lseek(fd, 0, SEEK_END);
	if (tar_len < 0) {
		LOGL(LOG_SSENGINE, "can't read tar_len (%s).\n", tar);
		goto Cleanup;
	}
	pos = lseek(fd, 0, SEEK_SET);
	if (pos < 0) {
		LOGL(LOG_SSENGINE, "can't read pos (%s).\n", tar);
		goto Cleanup;
	}
	while (pos < tar_len) {
		/* read file header */
		rdcnt = read(fd, header, sizeof(header));
		if (rdcnt <= 0) {
			LOG("read failed. (rdcnt=%d)\n", rdcnt);
			ret = -1;
			break;
		}
		/* get file name and file size */
		//memcpy(name, header, sizeof(name));
		if (header[TAR_ITEM_TYPE_FLAG_POS] == GNUTYPE_LONGNAME || header[TAR_ITEM_TYPE_FLAG_POS] == GNUTYPE_LONGLINK) {
			//rdcnt = read(fd, header, sizeof(header));
			memset(uExtendedName, 0, sizeof(uExtendedName));
			rdcnt = read(fd, uExtendedName, sizeof(uExtendedName) - 1);
			if (rdcnt <= 0) {
				LOG("read failed. (rdcnt=%d)\n", rdcnt);
				ret = -1;
				break;
			}
			rdcnt = read(fd, header, sizeof(header));
			if (rdcnt <= 0) {
				LOG("read failed. (rdcnt=%d)\n", rdcnt);
				ret = -1;
				break;
			}
		} else {
			memset(uExtendedName, 0, sizeof(uExtendedName));
			memcpy(uExtendedName, header, TAR_ITEM_NAME_SIZE);
		}
		memcpy(size_oct, header + TAR_ITEM_SIZE_POSITION, sizeof(size_oct));
		size_dec = strtoul(size_oct, NULL, TAR_SIZE_OF_ITEM_SIZE);
		if (size_dec > MAX_ITEM_SIZE) {
			LOG("size too big. (size_dec=0x%08X)\n", (unsigned int)size_dec);
			ret = -1;
			break;
		}
		//fix WGID : 51254 , size_dec comparison is not required
		if ((strstr(uExtendedName, "/diff") != NULL)) { //add only delta files from rootfs and csc, hardcoding shd b removed..

			/* check if the file is what we are looking for */
			//strncpy(itemName, name,100);
			itemSize = (int)size_dec;
			itemOffset = (int)lseek(fd, 0, SEEK_CUR);
			newnode = (tar_Data_t *) SS_Malloc(sizeof(tar_Data_t));
			if (!newnode) {
				ret = -1;
				break;
			}
			memset(newnode->itemName, 0, sizeof(newnode->itemName));
			strncpy((char *)newnode->itemName, uExtendedName, sizeof(newnode->itemName) - 1);
			newnode->itemOffset = itemOffset;
			newnode->itemSize = itemSize;
			newnode->nextnode = NULL;
			if (headparam == NULL) {
				headparam = newnode;
				tailparam = newnode;
			} else {
				(tailparam)->nextnode = newnode;
				(tailparam) = (tailparam)->nextnode;
			}
		}

		/* move file pointer to next file header */
		blknum = size_dec / TAR_BLOCK_SIZE;
		if (size_dec % TAR_BLOCK_SIZE)
			blknum++;

		pos = lseek(fd, (off_t) (blknum * TAR_BLOCK_SIZE), SEEK_CUR);
		if (pos < 0) {
			LOGL(LOG_SSENGINE, "can't read next block (%s).\n", tar);
			ret = -1;
			break;
		}
	}
 Cleanup:
	//if gTarFd was opened by tar_open during SS_FSUpdateFile we do not close it
	if (gTarFd < 0)
		close(fd);
	if (ret != -1)
		return headparam;
	else {
		if (headparam != newnode && newnode)
			SS_Free(newnode);
		if (headparam) {
			local_temp = headparam;
			while (local_temp) {
				local_next = local_temp->nextnode;
				SS_Free(local_temp);
				local_temp = local_next;
			}
		}
		return NULL;
	}
}

void tar_free_cfg_table(tar_Data_t ** delta_tar)
{
	tar_Data_t *local_temp = NULL;
	tar_Data_t *local_next = NULL;
	LOGL(LOG_SSENGINE, "Free TAR CFG TABLE\n");
	if (*delta_tar) {
		local_temp = *delta_tar;
		while (local_temp) {
			local_next = local_temp->nextnode;
			//LOGL(LOG_SSENGINE,"freeing [%s]\n",local_temp->itemName);
			SS_Free(local_temp);
			local_temp = local_next;
		}
	}
}

void deleteNode(tar_Data_t * head, tar_Data_t * n)
{
	tar_Data_t *prev = head;
	if (head == n) {
		if (head->nextnode == NULL) {
			LOG("There is only one node. The list can't be made empty ");
			return;
		}
		strncpy((char *)head->itemName, (const char *)head->nextnode->itemName, TAR_ITEM_NAME_SIZE);	//head->itemName = head->nextnode->itemName;
		head->itemSize = head->nextnode->itemSize;
		head->itemOffset = head->nextnode->itemOffset;
		n = head->nextnode;
		head->nextnode = head->nextnode->nextnode;
		SS_Free(n);
		return;
	}
	while (prev->nextnode != NULL && prev->nextnode != n)
		prev = prev->nextnode;
	if (prev->nextnode == NULL) {
		LOG("\n Given node is not present in Linked List");
		return;
	}
	prev->nextnode = prev->nextnode->nextnode;
	SS_Free(n);
	return;
}

int tar_get_item_size_from_struct(tar_Data_t ** delta_tar, const char *patchname, int *data_size, int *data_offset)
{
	tar_Data_t *head = *delta_tar;
	tar_Data_t *base = *delta_tar;
	if (head == NULL)
		return 1;
	else {
		//LOG("fast_tar_get_item_size- looking for [%s] [%s]\n",patchname,head->itemName);
		while (1) {
			if (strstr((const char *)head->itemName, patchname) != 0) {
				//LOG("fast_tar_get_item_size found [%s]  in [%s]\n",patchname, head->itemName);
				*data_size = head->itemSize;
				*data_offset = head->itemOffset;
				deleteNode(base, head);
				return 0;

			} else if (head->nextnode != NULL) {
				head = head->nextnode;
				//LOG("fast_tar_get_item_size current node [%s] \n",head->itemName);
			} else {
				LOGE("fast_tar_get_item_size FAILED TO GET [%s]  in [%s]\n", patchname, (char *)head->itemName);
				break;
			}
		}
		return 1;
	}
}

tar_Data_t *tar_cfg_clear_nodes(tar_Data_t * head)
{
	tar_Data_t *local_temp = NULL;
	while (head) {
		LOGL(LOG_SSENGINE, "tar_cfg_delete_node [%s]", (char *)head->itemName);
		local_temp = head->nextnode;
		SS_Free(head);
		head = local_temp;
	}
	return 0;
}

int tar_open(char *tar)
{
	if (gTarFd)
		close(gTarFd);
	gTarFd = open(tar, O_RDONLY);
	if (gTarFd < 0) {
		LOG("can't open TAR file(%s).\n", tar);
		return -1;
	}
	return 0;
}

int tar_close()
{
	if (gTarFd)
		close(gTarFd);
	gTarFd = -1;
	return 0;
}

int tar_get_folder_size(char *tar, char *item)
{
	int ret = -1;
	int fd = -1;
	char header[TAR_BLOCK_SIZE] = { 0, };
	char name[TAR_LONG_NAME_SIZE + 1] = { 0, };
	char *lastfolder = NULL;
	int folderpathlen = 0;
	char size_oct[TAR_SIZE_OF_HEADER] = { 0, };
	unsigned long size_dec = 0;
	int blknum = 0;
	off_t pos = 0;
	off_t tar_len = 0;
	ssize_t rdcnt = 0;
	LOG("Tar folder Looking for (%s)\n", item);
	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOG("can't open file(%s).\n", tar);
		return -1;
	}

	tar_len = lseek(fd, 0, SEEK_END);
	if (tar_len < 0) {
		LOGL(LOG_SSENGINE, "can't read tar_len (%s).\n", tar);
		goto Cleanup;
	}
	pos = lseek(fd, 0, SEEK_SET);
	if (pos < 0) {
		LOGL(LOG_SSENGINE, "can't read pos (%s).\n", tar);
		goto Cleanup;
	}

	while (pos < tar_len) {
		/* read file header */
		rdcnt = read(fd, header, sizeof(header));
		if (rdcnt <= 0) {
			LOG("read failed. (rdcnt=%d)\n", rdcnt);
			ret = -1;
			break;
		}

		/* get file name and file size */
		memcpy(name, header, sizeof(name) - 1);
		memcpy(size_oct, header + TAR_ITEM_SIZE_POSITION, sizeof(size_oct));
		size_dec = strtoul(size_oct, NULL, TAR_SIZE_OF_ITEM_SIZE);
		if (size_dec > MAX_ITEM_SIZE) {
			LOG("size too big. (size_dec=0x%08X)\n", (unsigned int)size_dec);
			ret = -1;
			break;
		}

		/* check if the file is what we are looking for */
		//Get until folder name

		lastfolder = strrchr(name, '/');
		if (lastfolder)
			folderpathlen = strlen(name) - strlen(lastfolder);

		if (strncmp(name, item, folderpathlen) == 0) {
			ret += (int)size_dec;
			//LOG("Tar Files under folder [%s]\n", name);
			//break;
		}

		/* move file pointer to next file header */
		blknum = size_dec / TAR_BLOCK_SIZE;
		if (size_dec % TAR_BLOCK_SIZE)
			blknum++;

		pos = lseek(fd, (off_t) (blknum * TAR_BLOCK_SIZE), SEEK_CUR);
		if (pos < 0) {
			LOGL(LOG_SSENGINE, "can't read next block (%s).\n", tar);
			close(fd);
			return -1;
		}
	}

 Cleanup:
	close(fd);
	LOG("ret=%d\n", ret);

	return ret;				 //Should return +1?? or Ignore??
}

/*Extract Specific Folder from tar, Taken from Untar.c */
int tar_extract_folder(char *tar, char *item, char *path)
{
	char buff[MAX_FILE_PATH];
	FILE *f = NULL;
	size_t bytes_read;
	int filesize;
	int data_offset = -1;
	int fd = -1;
	char name[512] = { 0, };
	int folderpathlen = 0;
	char dirPath[512] = { 0 };
	int getheader = 1;		  // Asuming initial header is TAR header
	char fullname[512] = { 0 };
	int ret;
	LOG("Extracting Folder from %s %s to %s\n", tar, item, path);

	data_offset = tar_get_item_offset(tar, item);
	if (data_offset < 0) {
		LOGE("data offset for [%s] is [%d]\n", item, data_offset);
		return -1;
	}

	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOGE("can't open file(%s).\n", tar);
		return -1;
	}

	folderpathlen = strlen(item);

	for (;;) {
		bytes_read = read(fd, buff, sizeof(buff));
		if (bytes_read < 512) {
			LOGE("Short read on %s: expected 512, got %d\n", tar, bytes_read);
			close(fd);
			return -1;
		}
		if (is_end_of_archive(buff)) {
			close(fd);
			LOG("End of %s\n", tar);	//Can stop at end of folder.
			return S_SS_SUCCESS;
		}
		if (!verify_checksum(buff)) {
			close(fd);
			LOGE("Checksum failure\n");
			return -1;
		}
		filesize = parseoct(buff + 124, 12);
		if (getheader == 2) {
			getheader = 1;
			//LOG(" Working on LONG FILE NAME CASE [%s]\n", fullname);
		} else {
			memset(fullname, 0, sizeof(fullname));
			strncpy(fullname, buff, 100);
			//LOG(" Working on Normal FILE NAME CASE [%s]\n", fullname);
		}

		switch (buff[156]) {
		case '1':
			LOG(" Ignoring hardlink %s\n", fullname);
			break;
		case '2':

			//LOG(" Creating symlink %s\n", buff);
			if (strncmp(fullname, item, folderpathlen) == 0) {
				//LOG("Printing Buffer \n");
				//for(i=157; buff[i] !='\0' ;i++)
				//{
				//LOG("%c", buff[i])	;
				//}
				//LOG("\nEnd buffer\n");
				memset(name, 0, sizeof(name));
				strncpy(name, buff + 157, 100); //157, target link name will be present
				memset(dirPath, 0, sizeof(dirPath));
				snprintf(dirPath, sizeof(dirPath), "%s/%s", path, fullname + folderpathlen);
				LOG(" Creating Symlink [%s][%s]\n", name, dirPath);
				ret = symlink(name, dirPath); // use ss_link
				if (ret < 0)
					LOG("Error with symlink: %d\n", errno);
			}
			break;
		case '3':
			LOG(" Ignoring character device %s\n", fullname);
			break;
		case '4':
			LOG(" Ignoring block device %s\n", fullname);
			break;
		case '5':
			//break;//delete
			//LOG(" Dir [%s] Item [%s] Length [%d]\n", fullname, item, folderpathlen);
			if (strncmp(fullname, item, folderpathlen) == 0) {
				//LOG(" Extracting dir %s\n", fullname);
				memset(dirPath, 0, sizeof(dirPath));
				snprintf(dirPath, sizeof(dirPath), "%s/%s", path, fullname + folderpathlen);
				create_dir(dirPath, parseoct(fullname + 100, 8));
			}

			filesize = 0;
			break;
		case '6':
			LOG(" Ignoring FIFO %s\n", fullname);
			break;
		case GNUTYPE_LONGLINK:
		case GNUTYPE_LONGNAME:
			{
				getheader = 2;
				memset(fullname, 0, sizeof(fullname));
				bytes_read = read(fd, fullname, sizeof(fullname));
				if (bytes_read < 512) {
					LOGE("Short read on %s: expected 512, got %d\n", tar, bytes_read);
					close(fd);
					return -1;
				}
				filesize = 0;
				//LOG("Entered LONG FILE NAME CASE  new NAME is [%s]\n", fullname);
				break;
			}

		default:
			//break;
			//LOG(" File [%s] Item [%s] Length [%d]\n", fullname, item, folderpathlen);
			if (strncmp(fullname, item, folderpathlen) == 0) {
				if (buff[PREFIX_INDICATOR_BYTE] != 0) {
					memset(name, 0, sizeof(name));
					memset(dirPath, 0, sizeof(dirPath));
					strncpy(name, buff, 100);
					snprintf(name + strlen(name), sizeof(name) - strlen(name),
							"%s", buff + PREFIX_INDICATOR_BYTE);
					snprintf(dirPath, sizeof(dirPath), "%s/%s", path, name + folderpathlen);
					LOG(" File Name is longer than 100 bytes -Remaining Str [%s]\n Full Str[%s]", dirPath);
				} else {
					//LOG(" Extracting file %s\n", fullname);
					memset(dirPath, 0, sizeof(dirPath));
					snprintf(dirPath, sizeof(dirPath), "%s/%s", path, fullname + folderpathlen);
					f = create_file(dirPath);
				}
			}

			break;
		}

		while (filesize > 0) {
			bytes_read = read(fd, buff, sizeof(buff));
			if (bytes_read < 512) {
				LOGE("Short read on %s: Expected 512, got %d\n", tar, bytes_read);
				close(fd);
				if (f) {
					fclose(f);//wgid: 16892
					f = NULL;
				}
				return -1;
			}
			if (filesize < 512)
				bytes_read = filesize;
			if (f != NULL) {
				if (fwrite(buff, 1, bytes_read, f)
					!= bytes_read) {
					LOG("Failed write\n");
					fclose(f);
					f = NULL;
					close(fd);//wgid: 59268
					return -1;//wgid: 16892
				}
			}
			filesize -= bytes_read;
		}
		if (f != NULL) {
			fclose(f);
			f = NULL;
		}

	}
	close(fd);
	return S_SS_SUCCESS;
}

int fast_tar_extract_file(char *tar, char *item, char *pathname, int size, int offset)
{
	int fd = -1;
	int data_size = size;
	int data_offset = offset;
	off_t pos = 0;
	ssize_t rdcnt = 0;
	ssize_t writeCount = 0;
	char *buf = NULL;
	char err_buf[256];
	int fd2;

	if (gTarFd > 0)
		fd = gTarFd;
	if (fd < 0) {
		fd = open(tar, O_RDONLY);
		if (fd < 0) {
			LOG("can't open file(%s).\n", tar);
			return -1;
		}
	}
	pos = lseek(fd, data_offset, SEEK_SET);
	if (pos < 0) {
		LOG("lseek fail (%s offset %d).\n", tar, data_offset);
		close(fd);
		return -1;
	}
	buf = SS_Malloc(data_size + 1);
	if (buf == NULL) {
		close(fd);
		LOGE("Failed to Allocate Memory\n");
		return -1;
	}
	rdcnt = read(fd, buf, data_size);
	if (rdcnt != (ssize_t) data_size) {
		LOG(" rdcnt read fail(%s from %s).\n", item, tar);
		SS_Free(buf);
		close(fd);
		return -1;
	}
	fd2 = open(pathname, O_CREAT | O_WRONLY, S_IRWXU);	 // Directory where file is required should be created already.
	if (fd2 < 0) {
		LOG("can't open file(%s).\n", pathname);
		SS_Free(buf);
		close(fd);
		return -1;
	}
	writeCount = write(fd2, buf, rdcnt);
	if (writeCount != rdcnt) {
		LOG("writeCount write fail(%s from %s).\n", item, tar);
		strerror_r(errno, err_buf, sizeof(err_buf));
		LOG("Oh  dear, something went wrong with read()! %s\n", err_buf);
		close(fd);
		close(fd2);
		SS_Free(buf);
		return -1;
	}
	SS_Free(buf);
	if (gTarFd < 0)
		close(fd);
	fsync(fd2);
	close(fd2);
	return rdcnt;			   // or jus return success?
}

int tar_extract_file(char *tar, char *item, char *pathname)
{
	int fd = -1;
	int data_size = -1;
	int data_offset = -1;
	off_t pos = 0;
	ssize_t rdcnt = 0;
	ssize_t writeCount = 0;
	char *buf = NULL;
	char err_buf[256];
	int fd2;
	data_size = tar_get_item_size(tar, item);
	data_offset = tar_get_item_offset(tar, item);

	if (data_size <= 0 || data_offset < 0) {
		LOGE("Error Not a file , size is [%d], offset [%d] for item [%s]\n", data_size, data_offset, item);
		return -1;
	} else
		LOGL(LOG_SSENGINE, "extracting file [%s] size [%d]\n", item, data_size);
	fd = open(tar, O_RDONLY);
	if (fd < 0) {
		LOG("can't open file(%s).\n", tar);
		return -1;
	}
	pos = lseek(fd, data_offset, SEEK_SET);
	if (pos < 0) {
		LOG("lseek fail (%s offset %d).\n", tar, data_offset);
		close(fd);
		return -1;
	}
	buf = SS_Malloc(data_size + 1);
	if (buf == NULL) {
		close(fd);
		LOGE("Failed to Allocate Memory\n");
		return -1;
	}
	rdcnt = read(fd, buf, data_size);
	if (rdcnt != (ssize_t) data_size) {
		LOG(" rdcnt read fail(%s from %s).\n", item, tar);
		SS_Free(buf);
		close(fd);
		return -1;
	}
	fd2 = open(pathname, O_CREAT | O_WRONLY, S_IRWXU);	 // Directory where file is required should be created already.
	if (fd2 < 0) {
		LOG("can't open file(%s).\n", pathname);
		SS_Free(buf);
		close(fd);
		return -1;
	}
	writeCount = write(fd2, buf, rdcnt);
	if (writeCount != rdcnt) {
		LOG("writeCount write fail(%s from %s).\n", item, tar);
		strerror_r(errno, err_buf, sizeof(err_buf));
		LOG("Oh  dear, something went wrong with read()! %s\n", err_buf);
		close(fd);
		close(fd2);
		SS_Free(buf);
		return -1;
	}
	SS_Free(buf);
	close(fd);
	fsync(fd2);
	close(fd2);
	return rdcnt;			   // or jus return success?
}
